#include "avisynth.h"
#include <vector>

class Feedback : public GenericVideoFilter {
public:
	struct FeedbackOutput {
		IScriptEnvironment* env;
		int id;
		PClip clip;
		bool linked;
	};

private:
	int id;
	bool output;
	int index = -1;
	PVideoFrame last = NULL;
	int last_n = -1;
	static std::vector<FeedbackOutput> outputs;

public:
	Feedback(PClip _child, int _id, bool _output, IScriptEnvironment* env);
	~Feedback();
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

std::vector<Feedback::FeedbackOutput> Feedback::outputs;

Feedback::Feedback(PClip _child, int _id, bool _output, IScriptEnvironment* env) : GenericVideoFilter(_child), id(_id), output(_output) {
	for (auto& o : outputs) {
		if (o.env == env && o.id == id) {
			auto _vi = o.clip->GetVideoInfo();
			if (_vi.width != vi.width || _vi.height != vi.height || _vi.pixel_type != vi.pixel_type) env->ThrowError("Feedback%s: Clip dimensions/colourspace mismatch", output ? "Output" : "Input");
			if (output) {
				if (o.linked) env->ThrowError("FeedbackOutput: ID already used");
				o.clip = this;
				o.linked = true;
			}
			return;
		}
	}

	outputs.push_back({ env, id, this, output });
}

Feedback::~Feedback() {
	last = NULL;
}

PVideoFrame __stdcall Feedback::GetFrame(int n, IScriptEnvironment* env) {
	if (output) {
		if (n != last_n) {
			last = child->GetFrame(n, env);
			last_n = n;
		}

		return last;
	} else {
		if (index == -1) {
			for (int i = 0; i < (int)outputs.size(); ++i) {
				if (outputs[i].env == env && outputs[i].id == id && outputs[i].linked) {
					index = i;
					break;
				}
			}

			if (index == -1) env->ThrowError("FeedbackInput: Output source not found");
		}

		if (n == 0) return child->GetFrame(n, env);
		return outputs[index].clip->GetFrame(n - 1, env);
	}
}

AVSValue __cdecl Create_Feedback(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new Feedback(
		args[0].AsClip(),
		args[1].AsInt(0),
		!!user_data,
		env
	);
}

const AVS_Linkage* AVS_linkage = 0;

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit3(IScriptEnvironment * env, const AVS_Linkage* const vectors) {
	AVS_linkage = vectors;
	env->AddFunction("FeedbackInput", "c[id]i", Create_Feedback, 0);
	env->AddFunction("FeedbackOutput", "c[id]i", Create_Feedback, (void*)1);

	return "Feedback";
}
